#version 450

in vec2 uv;

uniform sampler2DMS depth;
uniform sampler2DMS normals;

uniform mat4 inverse_view;
uniform mat4 inverse_projection;
uniform vec3 sun_light;

layout(r11f_g11f_b10f) uniform image3D lighting_map;

out vec3 color;

// Converts a color from linear light gamma to sRGB gamma
vec3 fromLinear(vec3 linearRGB)
{
    bvec3 cutoff = lessThan(linearRGB, vec3(0.0031308));
    vec3 higher = vec3(1.055)*pow(linearRGB, vec3(1.0/2.4)) - vec3(0.055);
    vec3 lower = linearRGB * vec3(12.92);

    return mix(higher, lower, cutoff);
}

/**
 * Generate a uniformly distributed random point on the unit-sphere.
 * 
 * After:
 * http://mathworld.wolfram.com/SpherePointPicking.html
 */
vec3 randomSpherePoint(vec3 rand) {
  float ang1 = (rand.x + 1.0) * 3.1415926535; // [-1..1) -> [0..2*PI)
  float u = rand.y; // [-1..1), cos and acos(2v-1) cancel each other out, so we arrive at [-1..1)
  float u2 = u * u;
  float sqrt1MinusU2 = sqrt(1.0 - u2);
  float x = sqrt1MinusU2 * cos(ang1);
  float y = sqrt1MinusU2 * sin(ang1);
  float z = u;
  return vec3(x, y, z);
}

/**
 * Generate a cosine-weighted random point on the unit hemisphere oriented around 'n'.
 * 
 * @param rand a vector containing pseudo-random values
 * @param n    the normal to orient the hemisphere around
 * @returns    the cosine-weighted point on the oriented hemisphere
 */
vec3 randomCosineWeightedHemispherePoint(vec3 rand, vec3 n) {
  float r = rand.x * 0.5 + 0.5; // [-1..1) -> [0..1)
  float angle = (rand.y + 1.0) * 3.1415926535; // [-1..1] -> [0..2*PI)
  float sr = sqrt(r);
  vec2 p = vec2(sr * cos(angle), sr * sin(angle));
  /*
   * Unproject disk point up onto hemisphere:
   * 1.0 == sqrt(x*x + y*y + z*z) -> z = sqrt(1.0 - x*x - y*y)
   */
  vec3 ph = vec3(p.xy, sqrt(1.0 - p*p));
  /*
   * Compute some arbitrary tangent space for orienting
   * our hemisphere 'ph' around the normal. We use the camera's up vector
   * to have some fix reference vector over the whole screen.
   */
  vec3 tangent = normalize(rand);
  vec3 bitangent = cross(tangent, n);
  tangent = cross(bitangent, n);
  
  /* Make our hemisphere orient around the normal. */
  return tangent * ph.x + bitangent * ph.y + n * ph.z;
}

highp float rand(vec2 co)
{
    highp float a = 12.9898;
    highp float b = 78.233;
    highp float c = 43758.5453;
    highp float dt= dot(co.xy ,vec2(a,b));
    highp float sn= mod(dt,3.14);
    return fract(sin(sn) * c);
}

void main() {
	vec3 collected_color = vec3(0., 0., 0.);
	for (int i = 0; i < gl_NumSamples; i++) {
		float depth = texelFetch(depth, ivec2(uv * textureSize(depth)), i).x;
		vec3 normal = texelFetch(normals, ivec2(uv * textureSize(normals)), i).xyz * 2 - 1.;

		vec4 position_w = inverse_view * inverse_projection * vec4(uv.xy * 2. - 1., depth * 2. - 1, 1.0);
		vec3 position = position_w.xyz / position_w.w;

		for (int i = 0; i < 24; i++) {
			ivec3 light_position = ivec3(position + imageSize(lighting_map).xyz / 2 + randomCosineWeightedHemispherePoint(vec3(rand(position.xy * (i + 1)), rand(position.xz * (i + 2)), rand(position.zy * (i + 3))) * 2. - 1., normal) * 0.5);

			if (any(lessThan(light_position, ivec3(0))) || any(greaterThanEqual(light_position, imageSize(lighting_map).xyz))) {
				collected_color += sun_light;
			} else {
				collected_color += imageLoad(lighting_map, light_position).rgb;
			}
		}
	}
	collected_color /= (24. * gl_NumSamples);

	color = collected_color;
}